import { EventEmitter } from '../../Events/EventEmitter.js';
import { EventFlowNode } from '../../Nodes/EventFlowNode.js';
import { Link } from '../../Nodes/Link.js';
import { sleep } from '../../sleep.js';
import { SyncExecutionBlock } from './SyncExecutionBlock.js';
export class GraphEvaluator {
    constructor(graph) {
        this.graph = graph;
        this.executionBlockQueue = [];
        this.asyncFlowNodes = [];
        this.eventFlowNodes = [];
        this.onNodeEvaluation = new EventEmitter();
        Object.values(this.graph.nodes).forEach((node) => {
            if (node instanceof EventFlowNode) {
                this.executionBlockQueue.push(new SyncExecutionBlock(this, new Link(node.id, '')));
            }
        });
    }
    asyncCommit(outputFlowSocket, syncEvaluationCompletedListener) {
        const node = this.graph.nodes[outputFlowSocket.nodeId];
        const outputSocket = node.outputSockets.find((socket) => socket.name === outputFlowSocket.socketName);
        if (outputSocket === undefined) {
            throw new Error(`no socket with the name ${outputFlowSocket.socketName}`);
        }
        if (outputSocket.links.length > 1) {
            throw new Error('invalid for an output flow socket to have multiple downstream links:' +
                `${node.description.typeName}.${outputSocket.name} has ${outputSocket.links.length} downlinks`);
        }
        if (outputSocket.links.length === 1) {
            const syncExecutionBlock = new SyncExecutionBlock(this, outputSocket.links[0], syncEvaluationCompletedListener);
            this.executionBlockQueue.push(syncExecutionBlock);
        }
    }
    executeAllSync(limitInSeconds = 100, limitInSteps = 100000000) {
        const startDateTime = Date.now();
        let elapsedSeconds = 0;
        let elapsedSteps = 0;
        while (elapsedSteps < limitInSteps &&
            elapsedSeconds < limitInSeconds &&
            this.executionBlockQueue.length > 0) {
            const currentExecutionBlock = this.executionBlockQueue[0];
            const localExecutionSteps = currentExecutionBlock.executeStep();
            if (localExecutionSteps < 0) {
                this.executionBlockQueue.shift();
            }
            elapsedSeconds = (Date.now() - startDateTime) * 0.001;
            if (localExecutionSteps > 0) {
                elapsedSteps += localExecutionSteps;
            }
        }
        return elapsedSteps;
    }
    async executeAllAsync(limitInSeconds = 100, limitInSteps = 100000000) {
        const startDateTime = Date.now();
        let elapsedSteps = 0;
        let elapsedTime = 0;
        let iterations = 0;
        do {
            if (iterations > 0) {
                await sleep(0);
            }
            elapsedSteps += this.executeAllSync(limitInSeconds - elapsedTime, limitInSteps - elapsedSteps);
            elapsedTime = (Date.now() - startDateTime) * 0.001;
            iterations += 1;
        } while ((this.asyncFlowNodes.length > 0 || this.executionBlockQueue.length > 0) &&
            elapsedTime < limitInSeconds &&
            elapsedSteps < limitInSteps);
        return elapsedSteps;
    }
}
//# sourceMappingURL=GraphEvaluator.js.map