import { Assert } from '../../Diagnostics/Assert.js';
import { FlowNode } from '../../Nodes/FlowNode.js';
import { ImmediateNode } from '../../Nodes/ImmediateNode.js';
import { Link } from '../../Nodes/Link.js';
import { NodeEvalContext } from '../../Nodes/NodeEvalContext.js';
import { NodeEvaluationEvent } from './NodeEvaluationEvent.js';
import { NodeEvaluationType } from './NodeEvaluationType.js';
export class SyncExecutionBlock {
    constructor(graphEvaluator, nextEval, syncEvaluationCompletedListener = undefined) {
        this.graphEvaluator = graphEvaluator;
        this.nextEval = nextEval;
        this.syncEvaluationCompletedListenerStack = [];
        this.graph = graphEvaluator.graph;
        if (syncEvaluationCompletedListener !== undefined) {
            this.syncEvaluationCompletedListenerStack.push(syncEvaluationCompletedListener);
        }
    }
    resolveInputValueFromSocket(inputSocket) {
        if (inputSocket.links.length === 0) {
            return 0;
        }
        const upstreamLink = inputSocket.links[0];
        const upstreamNode = this.graph.nodes[upstreamLink.nodeId];
        const upstreamOutputSocket = upstreamNode.outputSockets.find((socket) => socket.name === upstreamLink.socketName);
        if (upstreamOutputSocket === undefined) {
            throw new Error(`can not find socket with the name ${upstreamLink.socketName}`);
        }
        if (upstreamNode instanceof FlowNode) {
            inputSocket.value = upstreamOutputSocket.value;
            return 0;
        }
        if (!(upstreamNode instanceof ImmediateNode)) {
            throw new TypeError('node must be an instance of ImmediateNode');
        }
        let executionStepCount = 0;
        for (const upstreamInputSocket of upstreamNode.inputSockets) {
            executionStepCount +=
                this.resolveInputValueFromSocket(upstreamInputSocket);
        }
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(upstreamNode, NodeEvaluationType.Immediate, false));
        upstreamNode.evalFunc();
        executionStepCount++;
        this.graphEvaluator.onNodeEvaluation.emit(new NodeEvaluationEvent(upstreamNode, NodeEvaluationType.None, false));
        inputSocket.value = upstreamOutputSocket.value;
        return executionStepCount;
    }
    commit(outputFlowSocket, syncEvaluationCompletedListener = undefined) {
        Assert.mustBeTrue(this.nextEval === null);
        const node = this.graph.nodes[outputFlowSocket.nodeId];
        const outputSocket = node.outputSockets.find((socket) => socket.name === outputFlowSocket.socketName);
        if (outputSocket === undefined) {
            throw new Error(`can not find socket with the name ${outputFlowSocket.socketName}`);
        }
        if (outputSocket.links.length > 1) {
            throw new Error('invalid for an output flow socket to have multiple downstream links:' +
                `${node.description.typeName}.${outputSocket.name} has ${outputSocket.links.length} downlinks`);
        }
        if (outputSocket.links.length === 1) {
            const link = outputSocket.links[0];
            if (link === undefined) {
                throw new Error('link must be defined');
            }
            this.nextEval = link;
        }
        if (syncEvaluationCompletedListener !== undefined) {
            this.syncEvaluationCompletedListenerStack.push(syncEvaluationCompletedListener);
        }
    }
    executeStep() {
        const link = this.nextEval;
        this.nextEval = null;
        if (link === null) {
            if (this.syncEvaluationCompletedListenerStack.length === 0) {
                return -1;
            }
            const awaitingCallback = this.syncEvaluationCompletedListenerStack.pop();
            if (awaitingCallback === undefined) {
                throw new Error('awaitingCallback is empty');
            }
            awaitingCallback();
            return 0;
        }
        const node = this.graph.nodes[link.nodeId];
        let triggeringFlowSocket = undefined;
        let executionStepCount = 0;
        node.inputSockets.forEach((inputSocket) => {
            if (inputSocket.valueTypeName !== 'flow') {
                executionStepCount += this.resolveInputValueFromSocket(inputSocket);
            }
            else {
                if (inputSocket.name === link.socketName) {
                    inputSocket.value = true;
                    triggeringFlowSocket = inputSocket;
                }
                else {
                    inputSocket.value = false;
                }
            }
        });
        const context = new NodeEvalContext(this, node, triggeringFlowSocket);
        context.evalFlow();
        executionStepCount++;
        if (context.numCommits === 0 && !context.asyncPending) {
            let numFlowOutputs = 0;
            node.outputSockets.forEach((outputSocket) => {
                if (outputSocket.valueTypeName === 'flow') {
                    numFlowOutputs++;
                }
            });
            if (numFlowOutputs !== 1) {
                throw new Error(`can not use auto-commit if there are multiple flow outputs, number of outputs is ${numFlowOutputs} on ${node.description.typeName}`);
            }
            node.outputSockets.forEach((outputSocket) => {
                if (outputSocket.valueTypeName === 'flow') {
                    this.commit(new Link(link.nodeId, outputSocket.name));
                }
            });
        }
        return executionStepCount;
    }
}
//# sourceMappingURL=SyncExecutionBlock.js.map