import { Assert } from '../../Diagnostics/Assert.js';
import { AsyncNode } from '../../Nodes/AsyncNode.js';
import { FlowNode } from '../../Nodes/FlowNode.js';
import { ImmediateNode } from '../../Nodes/ImmediateNode.js';
export class Fiber {
    constructor(engine, nextEval, fiberCompletedListener = undefined) {
        this.engine = engine;
        this.nextEval = nextEval;
        this.fiberCompletedListenerStack = [];
        this.executionSteps = 0;
        this.graph = engine.graph;
        if (fiberCompletedListener !== undefined) {
            this.fiberCompletedListenerStack.push(fiberCompletedListener);
        }
    }
    resolveInputValueFromSocket(inputSocket) {
        if (inputSocket.links.length === 0) {
            return 0;
        }
        const upstreamLink = inputSocket.links[0];
        const upstreamNode = this.graph.nodes[upstreamLink.nodeId];
        const upstreamOutputSocket = upstreamNode.outputSockets.find((socket) => socket.name === upstreamLink.socketName);
        if (upstreamOutputSocket === undefined) {
            throw new Error(`can not find socket with the name ${upstreamLink.socketName}`);
        }
        if (upstreamNode instanceof FlowNode) {
            inputSocket.value = upstreamOutputSocket.value;
            return 0;
        }
        if (!(upstreamNode instanceof ImmediateNode)) {
            throw new TypeError('node must be an instance of ImmediateNode');
        }
        for (const upstreamInputSocket of upstreamNode.inputSockets) {
            this.resolveInputValueFromSocket(upstreamInputSocket);
        }
        this.engine.onNodeExecution.emit(upstreamNode);
        upstreamNode.exec();
        this.executionSteps++;
        inputSocket.value = upstreamOutputSocket.value;
    }
    commit(node, outputSocketName, fiberCompletedListener = undefined) {
        Assert.mustBeTrue(node instanceof FlowNode);
        Assert.mustBeTrue(this.nextEval === null);
        const outputSocket = node.outputSockets.find((socket) => socket.name === outputSocketName);
        if (outputSocket === undefined) {
            throw new Error(`can not find socket with the name ${outputSocketName}`);
        }
        if (outputSocket.links.length > 1) {
            throw new Error('invalid for an output flow socket to have multiple downstream links:' +
                `${node.description.typeName}.${outputSocket.name} has ${outputSocket.links.length} downlinks`);
        }
        if (outputSocket.links.length === 1) {
            const link = outputSocket.links[0];
            if (link === undefined) {
                throw new Error('link must be defined');
            }
            this.nextEval = link;
        }
        if (fiberCompletedListener !== undefined) {
            this.fiberCompletedListenerStack.push(fiberCompletedListener);
        }
    }
    executeStep() {
        const link = this.nextEval;
        this.nextEval = null;
        if (link === null) {
            if (this.fiberCompletedListenerStack.length === 0) {
                return -1;
            }
            const awaitingCallback = this.fiberCompletedListenerStack.pop();
            if (awaitingCallback === undefined) {
                throw new Error('awaitingCallback is empty');
            }
            awaitingCallback();
            return 0;
        }
        const node = this.graph.nodes[link.nodeId];
        let triggeredSocketName = '';
        node.inputSockets.forEach((inputSocket) => {
            if (inputSocket.valueTypeName !== 'flow') {
                this.resolveInputValueFromSocket(inputSocket);
            }
            else {
                inputSocket.value = inputSocket.name === link.socketName;
                if (inputSocket.value) {
                    triggeredSocketName = inputSocket.name;
                }
            }
        });
        this.engine.onNodeExecution.emit(node);
        if (node instanceof AsyncNode) {
            this.engine.asyncNodes.push(node);
            node.triggered(this.engine, triggeredSocketName, () => {
                const index = this.engine.asyncNodes.indexOf(node);
                this.engine.asyncNodes.splice(index, 1);
                this.executionSteps++;
            });
        }
        else if (node instanceof FlowNode) {
            node.triggered(this, triggeredSocketName);
            this.executionSteps++;
        }
    }
    isCompleted() {
        return (this.fiberCompletedListenerStack.length === 0 && this.nextEval === null);
    }
}
//# sourceMappingURL=Fiber.js.map